#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains InstanceGroup class."""

from yandex_cloud_client.base import YandexCloudObject


class InstanceGroup(YandexCloudObject):
    """This object represents a instance group."""

    pass


#!/usr/bin/env python3
"""A library that provides a Python interface to the Yandex.Cloud REST API."""

from .base import YandexCloudObject

from .client import YandexCloudClient, ComputeClient, CertificateClient

from .cloud import Cloud
from .folder import Folder, FolderSpec

from .certificate import Certificate, CertificateContent, CertificateRequestSpec, Challenges, DnsChallenges, HttpChallenges

from .compute.disk import Disk, DiskSpec, AttachedDisk, AttachedDiskSpec
from .compute.image import Image
from .compute.instance_group import InstanceGroup
from .compute.instance import Instance, InstanceSpec, ResourcesSpec
from .compute.placement_group import PlacementGroup
from .compute.snapshot import Snapshot

from .iam.service_account import ServiceAccountAuth

from .utils.decorators import retry, log
from .utils.helpers import generate_instance_yaml_example, instance_dict_example
from .utils.request import Request
from .utils.response import Response

from .vpc.address import Address, OneToOneNat
from .vpc.network_interface import NetworkInterface

from .operation import Operation, OperationWait
from .zone import Zone
from .version import __version__

__author__ = 'akimstrong@yandex.ru'

__all__ = [
    'YandexCloudObject', 'YandexCloudClient', 'ComputeClient', 'CertificateClient', 'Disk',
    'Certificate', 'CertificateContent', 'CertificateRequestSpec', 'Challenges', 'DnsChallenges',
    'HttpChallenges', 'DiskSpec', 'AttachedDisk', 'AttachedDiskSpec', 'Image', 'InstanceGroup',
    'Instance', 'InstanceSpec', 'ResourcesSpec', 'Operation', 'OperationWait', 'PlacementGroup',
    'Snapshot', 'ServiceAccountAuth', 'retry', 'log', 'Request', 'Response', 'Address', 'OneToOneNat',
    'NetworkInterface', 'Zone', 'Cloud', 'Folder', '__version__', '__author__', 'FolderSpec'
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains ServiceAccount, ServiceAccountAuth class."""

import jwt
import time

from yandex_cloud_client.base import YandexCloudObject
from yandex_cloud_client.utils.endpoints import IAM_URL


class ServiceAccount(YandexCloudObject):
    def __init__(self):
        pass


class ServiceAccountAuth:
    """This class represents auth method for service account.

    Args:
      :service_account_key: dict (decoded from key.json)

    Methods:
      generate_jwt()

    """

    def __init__(self, service_account_key: dict):
        self._sa_key = service_account_key
        self._validate_sa_format()

    @property
    def id(self):
        """Validating service account key."""
        _id = self._sa_key.get('id')
        if _id is None:
            raise KeyError(f'Invalid service account key, missing: id')
        return _id

    @property
    def service_account_id(self):
        """Validating service account id."""
        _service_account_id = self._sa_key.get('service_account_id')
        if _service_account_id is None:
            raise KeyError(f'Invalid service account key, missing: service_account_id')
        return _service_account_id

    @property
    def key_algorithm(self):
        return self._sa_key.get('key_algorithm')

    @property
    def private_key(self):
        """Validating service account private key."""
        prefix_start = '-----BEGIN PRIVATE KEY-----'
        _private_key = self._sa_key.get('private_key')

        if _private_key is None:
            raise KeyError(f'Invalid service account key, missing: private_key')

        elif not _private_key.startswith(prefix_start):
            raise TypeError(f'Invalid private key format, required format: SHA-256')

        return _private_key

    @property
    def public_key(self):
        return self._sa_key.get('public_key')

    def _validate_sa_format(self):
        """Validating service account key format."""
        if not isinstance(self._sa_key, dict):
            message = f'Invalid service account key. Required format: dict, ' + \
                      f'but received: {type(self._sa_key)}'
            raise TypeError(message)

    def generate_jwt(self):
        """This method prepare data for request IAM token."""
        now = time.time()
        payload = {
            "iss": self.service_account_id,
            "aud": IAM_URL + "/iam/v1/tokens",
            "iat": now,
            "exp": now + 360,
        }

        headers = {
            "typ": "JWT",
            "alg": "PS256",
            "kid": self.id,
        }

        return jwt.encode(payload, self.private_key, algorithm='PS256', headers=headers)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains RouteTable class."""

from yandex_cloud_client.base import YandexCloudObject


class RouteTable(YandexCloudObject):
    def __init__(self):
        pass



#!/usr/bin/env python3
# pylint: disable=E0401, C0114

import subprocess
import sys
from typing import Optional

import certifi

from . import __version__ as client_ver


def _git_revision() -> Optional[str]:
    try:
        output = subprocess.check_output(["git", "describe", "--long", "--tags"], stderr=subprocess.STDOUT)
    except (subprocess.SubprocessError, OSError):
        return None
    return output.decode().strip()


def print_ver_info() -> None:
    git_revision = _git_revision()
    print('yandex-cloud-client {}'.format(client_ver) + (' ({})'.format(git_revision) if git_revision else ''))
    print('certifi {}'.format(certifi.__version__))  # type: ignore[attr-defined]
    print('Python {}'.format(sys.version.replace('\n', ' ')))


def main() -> None:
    print_ver_info()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains helper functions."""

import re
import yaml
import json
import logging
import datetime as dt

logger = logging.getLogger(__name__)


def disk_mode_converter(mode):
    modes = {
        'ro': 'READ_ONLY',
        'rw': 'READ_WRITE'
    }

    key = mode.lower()
    return modes.get(key, mode)


def convert_yaml_to_dict(data: yaml):
    with open(data, 'r') as infile:
        result = yaml.load(infile, Loader=yaml.Loader)
    infile.close()

    return result


def convert_camel_to_snake(text: str):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', text)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def universal_obj_hook(obj: [list, dict]):
    if isinstance(obj, list):
        result = []
        for i in obj:
            result.append(_object_hook(i))
        return result

    elif isinstance(obj, dict):
        return _object_hook(obj)

    return obj


def _object_hook(obj: dict):
    cleaned_object = {}
    for key, value in obj.items():
        key = convert_camel_to_snake(key.replace('-', '_'))

        if len(key) and key[0].isdigit():
            key = '_' + key

        cleaned_object.update({key: value})
    return cleaned_object


def human_readable_size(raw_bytes, granularity=2):
    result = []
    sizes = (
        ('TB', 1024**4),
        ('GB', 1024**3),
        ('MB', 1024**2),
        ('KB', 1024),
        ('B', 1)
    )
    if raw_bytes == 0:
        return 0
    else:
        for name, count in sizes:
            value = raw_bytes // count
            if value:
                raw_bytes -= value * count
                result.append(f'{value} {name}')
        return ', '.join(result[:granularity])


def from_timestamp(unixtime):
    readable = dt.datetime.utcfromtimestamp(unixtime).isoformat()
    return readable


def string_to_datetime(strtime):
    try:
        raw_time = strtime.split('.')[0]
        return dt.datetime.strptime(raw_time, '%Y-%m-%dT%H:%M:%S')
    except IndexError:
        return strtime
    except ValueError:
        return dt.datetime.strptime(strtime, '%Y-%m-%dT%H:%M:%Sz')
    except Exception as err:
        logger.debug(err)
        return strtime


def generate_instance_yaml_example(path=None):
    filename = 'instance_example.yaml'
    if path is not None:
        if path.endswith('/'):
            path.strip('/')
        filename = f'{path}/instance_example.yaml'

    with open(filename, 'w') as outfile:
        yaml.dump(instance_dict_example(), outfile, sort_keys=False)
    outfile.close()
    print('File "instance_example.yaml" saved.')


def instance_dict_example():
    example = {
        "folderId": "string",
        "name": "string",
        "description": "string",
        "labels": "object",
        "zoneId": "string",
        "platformId": "string",
        "resourcesSpec": {
            "memory": "string",
            "cores": "string",
            "coreFraction": "string",
            "gpus": "string"
        },
        "metadata": "object",
        "bootDiskSpec": {
            "mode": "string",
            "deviceName": "string",
            "autoDelete": False,
            # `bootDiskSpec` includes only one of the fields `diskSpec`, `diskId`
            "diskSpec": {
                "name": "string",
                "description": "string",
                "typeId": "string",
                "size": "string",
            # `bootDiskSpec.diskSpec` includes only one of the fields `imageId`, `snapshotId`
            "imageId": "string",
            "snapshotId": "string",
            # end of the list of possible fields`bootDiskSpec.diskSpec`
            },
            "diskId": "string",
            # end of the list of possible fields`bootDiskSpec`
        },
        "secondaryDiskSpecs": [
            {
                "mode": "string",
                "deviceName": "string",
                "autoDelete": False,
                # `secondaryDiskSpecs[]` includes only one of the fields `diskSpec`, `diskId`
                "diskSpec": {
                    "name": "string",
                    "description": "string",
                    "typeId": "string",
                    "size": "string",
                    # `secondaryDiskSpecs[].diskSpec` includes only one of the fields `imageId`, `snapshotId`
                    "imageId": "string",
                    "snapshotId": "string",
                    # end of the list of possible fields`secondaryDiskSpecs[].diskSpec`
                },
            "diskId": "string",
            # end of the list of possible fields`secondaryDiskSpecs[]`
            }
        ],
        "networkInterfaceSpecs": [
            {
                "subnetId": "string",
                "primaryV4AddressSpec": {
                    "address": "string",
                    "oneToOneNatSpec": {
                        "ipVersion": "string",
                        "address": "string"
                    }
                },
                "primaryV6AddressSpec": {
                    "address": "string",
                    "oneToOneNatSpec": {
                        "ipVersion": "string",
                        "address": "string"
                    }
                }
            }
        ],
        "hostname": "string",
        "schedulingPolicy": {
            "preemptible": False
        },
        "serviceAccountId": "string",
        "networkSettings": {
            "type": "string"
        },
        "placementPolicy": {
            "placementGroupId": "string"
        }
    }

    return example



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains Snapshot class."""

from datetime import datetime

from yandex_cloud_client.constants import SECONDS_IN_DAY
from yandex_cloud_client.utils.helpers import string_to_datetime, human_readable_size
from yandex_cloud_client.base import YandexCloudObject


class Snapshot(YandexCloudObject):
    """This object represents a snapshot of the disk.

    Attributes:
      :id: str
      :folder_id: str
      :created_at: datetime
      :name: str
      :description: str
      :labels: list
      :storage_size: int
      :disk_size: int
      :product_ids: list
      :status: str
      :source_disk_id: str
      :client: object

    """

    def __init__(self,
                 id=None,
                 folder_id=None,
                 created_at=None,
                 name=None,
                 description=None,
                 labels=None,
                 storage_size=None,
                 disk_size=None,
                 product_ids=None,
                 status=None,
                 source_disk_id=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        # Required
        self.id = id

        # Optional
        self.folder_id = folder_id
        self.created_at = string_to_datetime(created_at) if created_at is not None else created_at
        self.name = name
        self.description = description
        self.labels = labels
        self.storage_size = int(storage_size) if storage_size is not None else storage_size
        self.disk_size = int(disk_size) if disk_size is not None else disk_size
        self.product_ids = product_ids
        self.status = status
        self.source_disk_id = source_disk_id

        self.client = client
        self._id_attrs = (self.id,)

    @property
    def age(self):
        """Snapshot age in days."""
        today = datetime.utcnow()
        if isinstance(self.created_at, datetime):
            _age = int((today - self.created_at).total_seconds()) // SECONDS_IN_DAY
            return _age
        return

    @property
    def human_readable_storage_size(self):
        """Converting bytes to human-readable size."""
        if self.storage_size is not None:
            return human_readable_size(self.storage_size)
        return self.storage_size

    @property
    def human_readable_disk_size(self):
        """Converting bytes to human-readable size."""
        if self.disk_size is not None:
            return human_readable_size(self.disk_size)
        return self.disk_size

    def delete(self, await_complete=True, run_async_await=False, *args, **kwargs):
        """Shortcut for client.delete_snapshot()."""
        return self.client.delete_snapshot(self.folder_id, self.id, await_complete=await_complete,
            run_async_await=run_async_await, *args, **kwargs)

    def update(self, *args, **kwargs):
        """Shortcut for client.update_snapshot()."""
        pass

    def operations(self, page_size=1000, *args, **kwargs):
        """Shortcut for snapshot_operations()."""
        return self.client.snapshot_operations(self.id, page_size=page_size, *args, **kwargs)

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(Snapshot, cls).de_json(data, client)
        return cls(client=client, **data)

    @classmethod
    def de_list(cls, data: list, client):
        if not data:
            return []

        snapshots = list()
        for snapshot in data:
            snapshots.append(cls.de_json(snapshot, client))

        return snapshots



class SnapshotSpec(YandexCloudObject):
    """This object represents a new snapshot."""

    def __init__(self,
                 folder_id=None,
                 disk_id=None,
                 name=None,
                 description=None,
                 labels=None,
                 client=None,
                 **kwargs):

        self.folderId = folder_id
        self.diskId = disk_id
        self.name = name
        self.description = description
        self.labels = labels

        self.client = client

    @classmethod
    def prepare(cls, data: dict, client):
        if not data:
            return None

        data = super(SnapshotSpec, cls).de_json(data, client)
        return cls(client=client, **data).to_clean_dict()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains Folder class."""

from yandex_cloud_client.base import YandexCloudObject
from yandex_cloud_client.utils.helpers import string_to_datetime


class Folder(YandexCloudObject):
    """This object represents a Folder."""

    def __init__(self,
                 id=None,
                 cloud_id=None,
                 created_at=None,
                 name=None,
                 description=None,
                 labels=None,
                 status=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.id = id
        self.cloud_id = cloud_id
        self.created_at = string_to_datetime(created_at) if created_at is not None else created_at
        self.name = name
        self.description = description
        self.labels = labels
        self.status = status

        self.client = client
        self._id_attrs = (self.id, self.cloud_id)

    def update(self):
        pass

    def delete(self):
        pass

    def operations(self, *args, **kwargs):
        """Shortcut for client.folder_operations()."""
        return self.client.folder_operations(self.id, *args, **kwargs)

    def show_access_bindings(self):
        pass

    def set_access_bindings(self):
        pass

    def update_access_bindings(self):
        pass

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(Folder, cls).de_json(data, client)
        return cls(client=client, **data)

    @classmethod
    def de_list(cls, data: list, client):
        if not data:
            return []

        folders = list()
        for folder in data:
            folders.append(cls.de_json(folder, client))

        return folders


class FolderSpec(YandexCloudObject):
    """This object represents a specification for new folder."""

    def __init__(self,
                 cloud_id=None,
                 name=None,
                 description=None,
                 labels=None,
                 client=None,
                 **kwargs):

        self.cloudId = cloud_id
        self.name = name
        self.description = description
        self.labels = labels

        self.client = client

    @classmethod
    def prepare(cls, data: dict, client):
        """Deserializing and preparing for a request."""
        if not data:
            return None

        data = super(FolderSpec, cls).de_json(data, client)
        result = cls(client=client, **data).to_dict()

        # cleaning None-type keys
        cleaner = lambda x: dict((k, v) for (k, v) in x.items() if v is not None)

        return cleaner(result)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains Operation and OperationWait classes."""

import time
import asyncio
import logging

from yandex_cloud_client.base import YandexCloudObject
from yandex_cloud_client.error import OperationDeadlineExceeded
from yandex_cloud_client.utils.helpers import string_to_datetime, universal_obj_hook
from yandex_cloud_client.utils.request import RpcError

logger = logging.getLogger(__name__)


class Operation(YandexCloudObject):
    """This object represents an operation."""

    def __init__(self,
                 id=None,
                 created_at=None,
                 created_by=None,
                 modified_at=None,
                 done=None,
                 response=None,
                 error=None,
                 metadata=None,
                 description=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        # Required
        self.id = id

        # Optional
        self.created_at = string_to_datetime(created_at) if created_at is not None else created_at
        self.created_by = created_by
        self.modified_at = string_to_datetime(modified_at) if modified_at is not None else modified_at
        self.done = bool(done)
        self.response = response
        self.error = error
        self.metadata = metadata
        self.description = description

        self.client = client
        self._id_attrs = (self.id,)

    @property
    def failed(self):
        if self.error is not None:
            return True
        return False

    @property
    def completed(self):
        if self.done:
            return True
        return False

    def cancel(self, *args, **kwargs):
        """Shortcut for client.cancel_operation()."""
        return self.client.cancel_operation(self.id, *args, **kwargs)

    def update_status(self, *args, **kwargs):
        """Shortcut for client.operation()."""
        return self.client.operation(self.id, *args, **kwargs)

    @classmethod
    def de_json(cls, data, client):
        if not data:
            return None

        data = super(Operation, cls).de_json(data, client)
        data['metadata'] = OperationMetadata.de_json(universal_obj_hook(data.get('metadata')), client)
        data['error'] = OperationError.de_json(universal_obj_hook(data.get('error')), client)
        return cls(client=client, **data)

    @classmethod
    def de_list(cls, data, client):
        if not data:
            return []

        operations = list()
        for operation in data:
            operations.append(cls.de_json(operation, client))

        return operations


class OperationMetadata(YandexCloudObject):
    """This object represents an operation metadata."""

    def __init__(self,
                 instance_id=None,
                 disk_id=None,
                 snapshot_id=None,
                 image_id=None,
                 subnet_id=None,
                 network_id=None,
                 certificate_id=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.instance_id = instance_id
        self.disk_id = disk_id
        self.snapshot_id = snapshot_id
        self.image_id = image_id
        self.subnet_id = subnet_id
        self.network_id = network_id
        self.certificate_id = certificate_id

        self.client = client

    @classmethod
    def de_json(cls, data, client):
        if not data:
            return None

        data = super(OperationMetadata, cls).de_json(data, client)
        return cls(client=client, **data)


class OperationError(YandexCloudObject):
    """This object represents an operation metadata."""

    def __init__(self,
                 code=None,
                 message=None,
                 details=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.code = RpcError(int(code)) if code is not None else code
        self.message = message
        self.details = details

        self.client = client

    @classmethod
    def de_json(cls, data, client):
        if not data:
            return None

        data = super(OperationError, cls).de_json(data, client)
        return cls(client=client, **data)



class OperationWait:
    """This class represents an operation waiter."""

    def __init__(self, operation: Operation, delay=2, timeout=600,
                 client=None, **kwargs):

        self._operation = operation
        self._delay = int(delay)
        self._deadline = time.time() + int(timeout) if timeout is not None else None

        self.client = client

    @property
    def operation(self):
        return self._operation

    @property
    def completed(self):
        """Sync operation waiter."""
        while not self._wait_operation():
            logger.debug('Sleeping.. sync awaiting operation status..')
            time.sleep(self._delay)
        return self._wait_operation()

    async def await_complete_async(self):
        """Async operation waiter."""
        while not self._wait_operation():
            logger.debug('Async awaiting operation status..')
            await asyncio.sleep(self._delay)
        return self._wait_operation()

    def _wait_operation(self):
        self._operation = self._operation.update_status()

        if self._operation.failed:
            logger.debug('Operation failed!')
            raise RuntimeError(self._operation.error)

        elif self._operation.completed:
            logger.debug('Operation done!')
            return self._operation

        elif time.time() >= self._deadline:
            logger.debug('Deadline exceeded!')
            raise OperationDeadlineExceeded(self._operation.id,
                self._operation.description)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains base endpoint urls."""

BASE_URL = 'https://api.cloud.yandex.net'
RESOURCE_MANAGER_URL = 'https://resource-manager.api.cloud.yandex.net'
IAM_URL = 'https://iam.api.cloud.yandex.net'
COMPUTE_URL = 'https://compute.api.cloud.yandex.net'
OPERATION_URL = 'https://operation.api.cloud.yandex.net'
CERTIFICATE_URL = 'https://certificate-manager.api.cloud.yandex.net'
CERTIFICATE_DATA_URL = 'https://data.certificate-manager.api.cloud.yandex.net'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains Cloud class."""

from yandex_cloud_client.base import YandexCloudObject
from yandex_cloud_client.utils.helpers import string_to_datetime


class Cloud(YandexCloudObject):
    """This object represents a Cloud."""

    def __init__(self,
                 id=None,
                 created_at=None,
                 name=None,
                 description=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.id = id
        self.created_at = string_to_datetime(created_at) if created_at is not None else created_at
        self.name = name
        self.description = description

        self.client = client
        self._id_attrs = (self.id,)

    def add_folder(self, *args, **kwargs):
        """Shortcut for client.create_folder()."""
        return self.client.create_folder(self.id, *args, **kwargs)

    def update(self):
        pass

    def operations(self, *args, **kwargs):
        """Shortcut for client.cloud_operations()."""
        return self.client.cloud_operations(self.id, *args, **kwargs)

    def folders(self, *args, **kwargs):
        """Shortcut for client.folders_in_cloud()."""
        return self.client.folders_in_cloud(self.id, *args, **kwargs)

    def show_access_bindings(self):
        pass

    def set_access_bindings(self):
        pass

    def update_access_bindings(self):
        pass

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(Cloud, cls).de_json(data, client)
        return cls(client=client, **data)

    @classmethod
    def de_list(cls, data: list, client):
        if not data:
            return []

        clouds = list()
        for cloud in data:
            clouds.append(cls.de_json(cloud, client))

        return clouds



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains Zone class."""

from yandex_cloud_client.base import YandexCloudObject


class Zone(YandexCloudObject):
    """This object represents the availability zone."""

    def __init__(self,
                 id=None,
                 region_id=None,
                 status=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.id = id
        self.region_id = region_id
        self.status = status

        self.client = client
        self._id_attrs = (self.id, self.region_id, self.status)


    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(Zone, cls).de_json(data, client)
        return cls(client=client, **data)

    @classmethod
    def de_list(cls, data: list, client):
        if not data:
            return []

        zones = list()
        for zone in data:
            zones.append(cls.de_json(zone, client))

        return zones



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains Image class."""

from yandex_cloud_client.base import YandexCloudObject


class Image(YandexCloudObject):
    """This object represents a image."""

    def __init__(self, client=None, **kwargs):
        pass

    def create(self, *args, **kwargs):
        pass

    def delete(self, *args, **kwargs):
        pass

    def update(self, *args, **kwargs):
        pass

    def operations(self, *args, **kwargs):
        pass



MIN_SHARED_INSTANCE_MEMORY = 524288
MIN_INSTANCE_MEMORY = 1048576
MAX_INSTANCE_MEMORY = 274877906944
MIN_DISK_SIZE = 4194304
MAX_DISK_SIZE = 4398046511104
INSTANCE_CORE_FRACTIONS = (0, 5, 20, 50, 100)
INSTANCE_CORES = (1,) + tuple(range(2, 34, 2))
INSTANCE_GPUS = (1,) + tuple(range(0, 6, 2))
DISK_MODES = ('READ_ONLY', 'READ_WRITE')
RESOURCE_ID_RE = '[a-z][a-z0-9-_]{,19}'
RESOURCE_NAME_RE = '|[a-z][-a-z0-9]{1,61}[a-z0-9]'
NETWORK_SETTINGS = ('STANDARD', 'SOFTWARE_ACCELERATED', 'HARDWARE_ACCELERATED')
INSTANCE_PLATFORMS = ('standard-v1', 'standard-v2')
IP_VERSIONS = ('IPV4', 'IPV6')
BASE_HEADERS = {'content-type': 'application/json'}
DEFAULT_TIMEOUT = 10
SECONDS_IN_DAY = 86400
AZ = ('ru-central1-a', 'ru-central1-b', 'ru-central1-c')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains Instance, Metadata, SchedulingPolicy, Resources classes."""

import logging

from yandex_cloud_client.utils.helpers import universal_obj_hook, human_readable_size, string_to_datetime
from yandex_cloud_client.base import YandexCloudObject
from yandex_cloud_client.compute.disk import AttachedDisk, DiskSpec
from yandex_cloud_client.vpc.network_interface import NetworkInterface
from yandex_cloud_client.error import TooManyArguments

from yandex_cloud_client.constants import (MAX_INSTANCE_MEMORY, MIN_SHARED_INSTANCE_MEMORY, 
                                           INSTANCE_CORES, INSTANCE_CORE_FRACTIONS, IP_VERSIONS,
                                           INSTANCE_PLATFORMS, AZ)

logger = logging.getLogger(__name__)


STATES = (
    'PROVISIONING',
    'RUNNING',
    'STOPPING',
    'STOPPED',
    'STARTING',
    'RESTARTING',
    'UPDATING',
    'ERROR',
    'CRASHED',
    'DELETING',
)


class Instance(YandexCloudObject):
    """This object represents a instance.

    Attributes:
      :id: str
      :folder_id: str
      :created_at: datetime
      :name: str
      :description: str
      :labels: list
      :zone_id: str
      :platform_id: str
      :resources: object
      :status: str
      :metadata: object
      :boot_disk: object
      :secondary_disk: list of obj
      :network_interfaces: list of obj
      :fqdn: str
      :scheduling_policy: object
      :service_account_id: str
      :network_settings: object
      :client: object

    """

    def __init__(self,
                 id=None,
                 folder_id=None,
                 created_at=None,
                 name=None,
                 description=None,
                 labels=None,
                 zone_id=None,
                 platform_id=None,
                 resources=None,
                 status=None,
                 metadata=None,
                 boot_disk=None,
                 secondary_disks=None,
                 network_interfaces=None,
                 fqdn=None,
                 scheduling_policy=None,
                 service_account_id=None,
                 network_settings=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        # Required
        self.id = id

        # Optional
        self.folder_id = folder_id
        self.created_at = string_to_datetime(created_at) if created_at is not None else created_at
        self.name = name
        self.description = description
        self.labels = labels
        self.zone_id = zone_id
        self.platform_id = platform_id
        self.resources = resources
        self.status = status
        self.metadata = metadata
        self.boot_disk = boot_disk
        self.secondary_disks = secondary_disks
        self.network_interfaces = network_interfaces
        self.fqdn = fqdn
        self.scheduling_policy = scheduling_policy
        self.service_account_id = service_account_id
        self.network_settings = network_settings

        self.client = client
        self._id_attrs = (self.id,)

    @property
    def running(self):
        if self.status == 'RUNNING':
            return True
        return False

    @property
    def stopped(self):
        if self.status == 'STOPPED':
            return True
        return False

    @property
    def crashed(self):
        if self.status == 'CRASHED':
            return True
        return False

    @property
    def error(self):
        if self.status == 'ERROR':
            return True
        return False

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(Instance, cls).de_json(data, client)
        data['boot_disk'].update({'attached_to': data.get('id', None)})
        data['boot_disk'].update({'folder_id': data.get('folder_id', None)})
        data['boot_disk'] = AttachedDisk.de_json(universal_obj_hook(data.get('boot_disk')), client)

        # add folder id and source instance id as attr to disks
        for disk in data.get('secondary_disks', []):
            if disk is not None:
                disk.update({'attached_to': data.get('id', None)})
                disk.update({'folder_id': data.get('folder_id', None)})

        data['secondary_disks'] = AttachedDisk.de_list(universal_obj_hook(data.get('secondary_disks')), client)
        data['network_interfaces'] = NetworkInterface.de_list(universal_obj_hook(data.get('network_interfaces')), client)
        data['resources'] = Resources.de_json(universal_obj_hook(data.get('resources')), client)
        data['metadata'] = Metadata.de_json(universal_obj_hook(data.get('metadata')), client)
        data['scheduling_policy'] = SchedulingPolicy.de_json(universal_obj_hook(data.get('scheduling_policy')), client)
        data['network_settings'] = NetworkSettings.de_json(universal_obj_hook(data.get('network_settings')), client)

        return cls(client=client, **data)

    @classmethod
    def de_list(cls, data: list, client):
        if not data:
            return []

        instances = list()
        for instance in data:
            instances.append(cls.de_json(instance, client))

        return instances

    def start(self, run_async_await=False, await_complete=True, *args, **kwargs):
        """Shortcut for client.start_instance()."""
        return self.client.start_instance(self.id, await_complete, run_async_await, *args, **kwargs)

    def restart(self, run_async_await=False, await_complete=True, *args, **kwargs):
        """Shortcut for client.restart_instance()."""
        return self.client.restart_instance(self.id, await_complete, run_async_await, *args, **kwargs)

    def stop(self, run_async_await=False, await_complete=True, *args, **kwargs):
        """Shortcut for client.stop_instance()."""
        return self.client.stop_instance(self.id, await_complete, run_async_await, *args, **kwargs)

    def delete(self, run_async_await=False, await_complete=True, *args, **kwargs):
        """Shortcut for client.delete_instance()."""
        return self.client.delete_instance(self.id, await_complete, run_async_await, *args, **kwargs)

    def update(self, run_async_await=False, await_complete=True, *args, **kwargs):
        """Shortcut for client.update_instance()."""
        pass

    def update_metadata(self, run_async_await=False, await_complete=True, *args, **kwargs):
        """Shortcut for client.update_instance_metadata()."""
        pass

    def operations(self, page_size=1000, *args, **kwargs):
        """Shortcut for client.instance_operations()."""
        return self.client.instance_operations(self.id, page_size=page_size, *args, **kwargs)

    def attach_new_disk(self, name=None, description=None,
                       type_id=None, size=None, image_id=None,
                       snapshot_id=None, *args, **kwargs):
        """Create new disk and attach to the instance."""
        if image_id and snapshot_id:
            raise ValueError('SOME MESSAGE')
        else:
            pass

    def attach_existent_disk(self, disk_id=None, *args, **kwargs):
        """Shortcut for Client.instance_attach_existent_disk()."""
        return self.client.instance_attach_existent_disk(self.id, disk_id=disk_id,
            await_complete=True, run_async_await=False, *args, **kwargs)

    def detach_disk(self, disk_id=None, disk_name=None, await_complete=True,
                    run_async_await=False, *args, **kwargs):
        """Shortcut for Client.instance_detach_disk()."""
        return self.client.instance_detach_disk(self.id, disk_id=disk_id, disk_name=disk_name,
            await_complete=await_complete, run_async_await=run_async_await, *args, **kwargs)

    def attached_disks(self, *args, **kwargs):
        """Shortcut for private method Client._convert_attached_disks()."""
        disks = [self.boot_disk.id] + [x.id for x in self.secondary_disks]
        return self.client._convert_attached_disks(disks, *args, **kwargs)

    def serial_port_output(self, port=None, *args, **kwargs):
        """Shortcut for client.instance_serial_port_output()."""
        return self.client.instance_serial_port_output(self.id, port=port, *args, **kwargs)


    # Aliases
    attachedDisks = attached_disks
    attachExistentDisk = attach_existent_disk
    attachNewDisk = attach_new_disk
    detachDisk = detach_disk
    updateMetadata = update_metadata
    serialPortOutput = serial_port_output


# Subsidiary classes for Instance object


class Metadata(YandexCloudObject):
    """This object represents a metadata of instance."""

    def __init__(self,
                 user_data=None,
                 serial_port_enable=None,
                 ssh_keys=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.user_data = user_data
        self.serial_port_enable = bool(int(serial_port_enable)) if serial_port_enable is not None else serial_port_enable
        self.ssh_keys = ssh_keys

        self.client = client
        self._id_attrs = (self.user_data, self.serial_port_enable, self.ssh_keys)

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(Metadata, cls).de_json(data, client)
        return cls(client=client, **data)


class SchedulingPolicy(YandexCloudObject):
    """This object represents a sheduling policy options of instance."""

    def __init__(self,
                 preemptible=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.preemptible = bool(preemptible)
        self.client = client
        self._id_attrs = (self.preemptible,)

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(SchedulingPolicy, cls).de_json(data, client)
        return cls(client=client, **data)


class Resources(YandexCloudObject):
    """This object represents a resources of instance."""

    def __init__(self, memory=None, cores=None, core_fraction=None,
                 gpus=None, client=None, **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.memory = int(memory) if memory is not None else memory
        self.cores = int(cores) if cores is not None else cores
        self.core_fraction = int(core_fraction) if core_fraction is not None else core_fraction
        self.gpus = int(gpus) if gpus is not None else gpus

        self.client = client
        self._id_attrs = (self.memory, self.cores, self.core_fraction, self.gpus)

    @property
    def human_readable_memory(self):
        if self.memory is not None:
            return human_readable_size(self.memory)
        return self.memory

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(Resources, cls).de_json(data, client)
        return cls(client=client, **data)


class NetworkSettings(YandexCloudObject):
    """This object represents a network settings of instance."""

    def __init__(self,
                 type=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.type = type
        self.client = client
        self._id_attrs = (self.type,)

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(NetworkSettings, cls).de_json(data, client)
        return cls(client=client, **data)


# Specifications for new instances

# FIXME: not completely done
class InstanceSpec(YandexCloudObject):
    """This object represents new instance spec."""

    def __init__(self,
                 folder_id=None,
                 name=None,
                 description=None,
                 labels=None,
                 zone_id=None,
                 platform_id=None,
                 resources_spec=None,
                 metadata=None,
                 boot_disk_spec=None,
                 secondary_disk_specs=None,
                 network_interface_specs=None,
                 hostname=None,
                 scheduling_policy=None,
                 service_account_id=None,
                 network_settings=None,
                 client=None,
                 **kwargs):

        self.folderId = folder_id
        self.name = name
        self.description = description
        self.labels = labels
        self.zoneId = zone_id
        self.platformId = platform_id
        self.resourcesSpec = resources_spec
        self.metadata = metadata
        self.bootDiskSpec = boot_disk_spec
        self.secondaryDiskSpecs = secondary_disk_specs
        self.networkInterfaceSpecs = network_interface_specs
        self.hostname = hostname
        self.schedulingPolicy = scheduling_policy
        self.serviceAccountId = service_account_id
        self.networkSettings = network_settings

        self.client = client


    @classmethod
    def prepare(cls, data: dict, client):
        """Deserializing and preparing for a request."""
        if not data:
            return None

        data = super(InstanceSpec, cls).de_json(data, client)
        data['resourcesSpec'] = ResourcesSpec.de_json(data.get('resourcesSpec'), client)
        # data['bootDiskSpec'] = None
        # data['secondaryDiskSpecs'] = None
        # data['networkInterfaceSpecs'] = None
        # data['schedulingPolicy'] = None
        # data['networkSettings'] = None
        return cls(client=client, **data)


class ResourcesSpec(YandexCloudObject):
    """This object represents a resources specification for new instance."""

    def __init__(self,
                 memory=None,
                 cores=None,
                 core_fraction=None,
                 gpus=None,
                 client=None,
                 **kwargs):

        self.memory = int(memory) if memory is not None else memory
        self.cores = int(cores) if cores is not None else cores
        self.coreFraction = int(core_fraction) if core_fraction is not None else core_fraction
        self.gpus = int(gpus) if gpus is not None else gpus

        self.client = client

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(ResourcesSpec, cls).de_json(data, client)
        return cls(client=client, **data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains Address, OneToOneNat and PrimaryV4,6Address classes."""
import logging

from yandex_cloud_client.base import YandexCloudObject
from yandex_cloud_client.utils.helpers import universal_obj_hook

logger = logging.getLogger(__name__)


class OneToOneNat(YandexCloudObject):
    """This object represents a one-to-one NAT address."""

    def __init__(self,
                 address=None,
                 ip_version=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.address = address
        self.ip_version = ip_version

        self.client = client
        self._id_attrs = (self.address, self.ip_version)

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(OneToOneNat, cls).de_json(data, client)
        return cls(client=client, **data)

    @classmethod
    def de_list(cls, data: list, client):
        if not data:
            return []

        nats = list()
        for nat in data:
            nats.append(cls.de_json(nat, client))

        return nats


class Address(YandexCloudObject):
    """This object represents a base network address."""

    def __init__(self,
                 address=None,
                 one_to_one_nat=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.address = address
        self.one_to_one_nat = one_to_one_nat

        self.client = client
        self._id_attrs = (self.address,)

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(Address, cls).de_json(data, client)
        data['one_to_one_nat'] = OneToOneNat.de_json(universal_obj_hook(data.get('one_to_one_nat')), client)
        return cls(client=client, **data)

    @classmethod
    def de_list(cls, data: list, client):
        if not data:
            return []

        addresses = list()
        for address in data:
            addresses.append(cls.de_json(address, client))

        return addresses



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains YandexCloudClient, ComputeClient classes"""

import json
import logging
import requests

from types import CoroutineType

from yandex_cloud_client.base import YandexCloudObject

from yandex_cloud_client.utils.request import Request
from yandex_cloud_client.utils.decorators import log
from yandex_cloud_client.utils.helpers import convert_yaml_to_dict
from yandex_cloud_client.utils.endpoints import (
    BASE_URL,
    IAM_URL,
    COMPUTE_URL,
    CERTIFICATE_URL,
    OPERATION_URL,
    RESOURCE_MANAGER_URL,
    CERTIFICATE_DATA_URL
)
from yandex_cloud_client.constants import BASE_HEADERS, DEFAULT_TIMEOUT
from yandex_cloud_client.error import (
    InvalidToken,
    YandexCloudError,
    TooManyArguments,
    BadRequest,
    MethodNotAvailable
)

from yandex_cloud_client.cloud import Cloud
from yandex_cloud_client.folder import Folder, FolderSpec

from yandex_cloud_client.iam.service_account import ServiceAccountAuth
from yandex_cloud_client.operation import Operation, OperationWait

from yandex_cloud_client.compute.disk import Disk, DiskSpec, AttachedDiskSpec
from yandex_cloud_client.compute.instance import Instance, InstanceSpec
from yandex_cloud_client.compute.snapshot import Snapshot, SnapshotSpec

from yandex_cloud_client.certificate import Certificate, CertificateRequestSpec, CertificateContent


logger = logging.getLogger(__name__)

# TODO (akimrx): add accessBindings for all clients
# TODO (akimrx): add update() methods for all resources
# TODO (akimrx): add pagintaion for list requests
# TODO (akimrx): add IAM methods to Base Client


class YandexCloudClient(YandexCloudObject):
    """Yandex.Cloud Client.
    Probably, this REST API Client will make your life with Yandex.Cloud a little easier.

    This parent Client contains authorization & general methods for services:
      - IAM
      - Operation
      - Resource Manager
      - Virtual Private Cloud

    Other services like a Compute, Managed Databases, etc — are inheritors of this client.
    For example, to work with the Compute service, use the `ComputeClient` class.
    It will already contain the methods of this client.

    Required Args:
      oauth_token: str
            or
      iam_token: str
            or
      service_account_key: dict

    Optional Args:
      request: Request
      base_url: str
      compute_url: str
      resource_manager_url: str
      iam_url: str
      operation_url: str
      certificate_url: str
      certificate_data_url: str

    Methods:
      operation()                     -> return Operation object
      operation_cancel()              -> return Operation object

      """

    def __init__(self,
                 oauth_token: str = None,
                 iam_token: str = None,
                 service_account_key: dict = None,
                 request: object = None,
                 timeout: int = None,
                 base_url: str = None,
                 resource_manager_url: str = None,
                 compute_url: str = None,
                 certificate_url: str = None,
                 certificate_data_url: str = None,
                 iam_url: str = None,
                 operation_url: str = None):

        _cred_args = [x for x in (service_account_key, oauth_token, iam_token) if x is not None]
        if len(_cred_args) > 1:
            message = f'Too many credentials({len(_cred_args)}) received, ' + \
                      'but only one credential type can be specified'
            raise TooManyArguments(message)
        elif len(_cred_args) < 1:
            raise InvalidToken('IAM/OAuth token or service account key required!')

        if oauth_token:
            self.token = YandexCloudClient.get_iam_token(oauth_token)
        elif service_account_key:
            self.token = YandexCloudClient.get_token_for_sa(service_account_key)
        elif iam_token:
            self.token = iam_token
        else:
            raise InvalidToken('IAM/OAuth token or service account key required!')

        self.timeout = int(timeout) if timeout is not None else DEFAULT_TIMEOUT

        if request:
            self._request = request
            self._request.set_and_return_client(self)
        else:
            self._request = Request(self, timeout=self.timeout)

        self.base_url = base_url or BASE_URL
        self.iam_url = iam_url or IAM_URL
        self.resource_manager_url = resource_manager_url or RESOURCE_MANAGER_URL
        self.compute_url = compute_url or COMPUTE_URL
        self.certificate_url = certificate_url or CERTIFICATE_URL
        self.certificate_data_url = certificate_data_url or CERTIFICATE_DATA_URL
        self.operation_url = operation_url or OPERATION_URL

    @staticmethod
    def get_iam_token(oauth_token, raw=False) -> [str, dict]:
        """Returns IAM Token for user account."""
        url = f'{IAM_URL}/iam/v1/tokens'
        data = {'yandexPassportOauthToken': oauth_token}

        result = requests.post(url, json=data, headers=BASE_HEADERS)
        try:
            response = result.json()
        except json.decoder.JSONDecodeError:
            raise BadRequest(result.text)

        if not result.ok:
            raise BadRequest(result.text)

        if raw:
            return response
        return response.get('iamToken')

    @staticmethod
    def get_token_for_sa(sa_credentials: dict, raw=False) -> [str, dict]:
        """Returns IAM Token for service account."""
        url = f'{IAM_URL}/iam/v1/tokens'
        data = {'jwt': ServiceAccountAuth(sa_credentials).generate_jwt()}

        result = requests.post(url, params=data, headers=BASE_HEADERS)
        try:
            response = result.json()
        except json.decoder.JSONDecodeError:
            raise BadRequest(result.text)

        if not result.ok:
            raise BadRequest(result.text)
        if raw:
            return response
        return response.get('iamToken')

    @staticmethod
    def get_endpoints_from_api() -> dict:
        """Return endpoints as dict."""
        url = f'{BASE_URL}/endpoints'
        result = requests.get(url)
        response = result.json()

        if not result.ok:
            raise BadRequest(result.text)

        return response.get('endpoints')

    # Private methods for ComputeClient workflow

    def _delete_resource(self, url, await_complete=None) -> Operation:
        """Wrapper for delete resources."""
        response = self._request.delete(url)
        operation = Operation.de_json(response, self)
        if not await_complete:
            return operation
        return OperationWait(operation).completed

    # actually, half-async (because Requests)
    async def _async_delete_resource(self, url) -> CoroutineType:
        """Half async func for delete resource.
        Returns coroutine.
        """
        response = self._request.delete(url)
        operation = Operation.de_json(response, self)

        await OperationWait(operation).await_complete_async()

    def _resource_create(self, url, data=None, await_complete=True) -> Operation:
        """Wrapper for create resource."""
        response = self._request.post(url, json=data)
        operation = Operation.de_json(response, self)

        if not await_complete:
            return operation
        return OperationWait(operation).completed

    # actually, half-async (because Requests)
    async def _async_resource_create(self, url, data=None) -> Operation:
        """Half async func for create resource.
        Returns coroutine.
        """
        response = self._request.post(url, json=data)
        operation = Operation.de_json(response, self)

        await OperationWait(operation).await_complete_async()

    # Operations public methods

    @log
    def operation(self, operation_id: str) -> Operation:
        """Returns operation as object."""
        url = f'{self.operation_url}/operations/{operation_id}'
        response = self._request.get(url)
        return Operation.de_json(response, self)

    @log
    def cancel_operation(self, operation_id: str) -> Operation:
        """Cancels the operation if possible."""
        url = f'{self.operation_url}/operations/{operation_id}:cancel'
        try:
            response = self._request.post(url)
            return Operation.de_json(response, self)
        except YandexCloudError:
            raise BadRequest("Operation can't be canceled.")

    @log
    def cloud(self, cloud_id: str) -> Cloud:
        """Returns cloud as object."""
        url = f'{self.resource_manager_url}/resource-manager/v1/clouds/{cloud_id}'

        response = self._request.get(url)
        return Cloud.de_json(response, self)

    @log
    def available_clouds(self, page_size: int = 1000, query_filter: str = None) -> [Cloud]:
        """Returns list of available clouds for account."""
        url = f'{self.resource_manager_url}/resource-manager/v1/clouds?pageSize={page_size}'

        if query_filter:
            url += f'&filter={query_filter}'

        response = self._request.get(url).get('clouds')
        return Cloud.de_list(response, self)

    @log
    def update_cloud(self):
        raise MethodNotAvailable(f'Method update_cloud is not support yet')

    @log
    def cloud_operations(self, cloud_id: str, page_size=1000) -> [Operation]:
        """Returns list of operations in the cloud."""
        url = f'{self.resource_manager_url}/resource-manager/v1/clouds/{cloud_id}/operations?pageSize={page_size}'
        response = self._request.get(url).get('operations')
        return Operation.de_list(response, self)

    @log
    def cloud_access_bindings(self):
        raise MethodNotAvailable(f'Method cloud_access_bindings is not support yet')

    @log
    def update_cloud_access_bindings(self):
        raise MethodNotAvailable(f'Method update_cloud_access_bindings is not support yet')

    @log
    def set_cloud_access_bindings(self):
        raise MethodNotAvailable(f'Method set_cloud_access_bindings is not support yet')

    @log
    def folder(self, folder_id: str) -> Folder:
        """Returns folder as object."""
        url = f'{self.resource_manager_url}/resource-manager/v1/folders/{folder_id}'

        response = self._request.get(url)
        return Folder.de_json(response, self)

    @log
    def folders_in_cloud(self, cloud_id: str, page_size: int = 1000, query_filter: str = None) -> [Folder]:
        """Returns all available folders for user / service account,
        depending on the type of authorization and permissions.
        """
        url = f'{self.resource_manager_url}/resource-manager/v1/folders?cloudId={cloud_id}&pageSize={page_size}'

        if query_filter:
            url += f'&filter={query_filter}'

        response = self._request.get(url)
        print(response)
        return Folder.de_list(response.get('folders'), self)

    @log
    def folder_operations(self, folder_id: str, page_size=1000) -> [Operation]:
        """Returns list of operations in the folder."""
        url = f'{self.resource_manager_url}/resource-manager/v1/folders/{folder_id}/operations?pageSize={page_size}'
        response = self._request.get(url).get('operations')
        return Operation.de_list(response, self)

    @log
    def folder_access_bindings(self):
        raise MethodNotAvailable(f'Method folder_access_bindings is not support yet')

    @log
    def update_folder_access_bindings(self):
        raise MethodNotAvailable(f'Method update_folder_access_bindings is not support yet')

    @log
    def set_folder_access_bindings(self):
        raise MethodNotAvailable(f'Method set_folder_access_bindings is not support yet')

    @log
    def create_folder(self,
                      cloud_id: str,
                      name: str,
                      await_complete=True,
                      run_async_await=False) -> Operation:
        """Create folder in the specified cloud."""
        url = f'{self.resource_manager_url}/resource-manager/v1/folders'

        params = locals().copy()
        del params['self']
        data = FolderSpec.prepare(params, self)

        if run_async_await:
            return self._async_resource_create(url, data=data)
        return self._resource_create(url, data=data, await_complete=await_complete)

    @log
    def update_folder(self):
        raise MethodNotAvailable(f'Method update_folder is not support yet')

    @log
    def delete_folder(self):
        raise MethodNotAvailable(f'Method delete_folder is not support yet')


    # Aliases

    getEndpoints = get_endpoints_from_api
    getIamToken = get_iam_token
    getServiceAccountToken = get_token_for_sa
    cancelOperation = cancel_operation


class ComputeClient(YandexCloudClient):
    """Yandex.Cloud Compute Client.

    Child class, which inherit the properties and methods
    from the YandexCloudClient parent class.

    Required Args:
      oauth_token: str
            or
      iam_token: str
            or
      service_account_key: dict

    Optional Args:
      request: Request
      base_url: str
      compute_url: str
      resource_manager_url: str
      iam_url: str
      operation_url: str
      certificate_url: str
      certificate_data_url: str

    """

    def _instance_state_management(self, action=None, instance_id=None,
                                   await_complete=None) -> Operation:
        """Helper func for managing the state of the instance.
        Supported actions: start, restart, stop.
        """

        ACTIONS = ('start', 'restart', 'stop')
        if action not in ACTIONS:
            raise TypeError(f'Action {action} not supported')

        url = f'{self.compute_url}/compute/v1/instances/{instance_id}:{action}'
        response = self._request.post(url)
        operation = Operation.de_json(response, self)
        if not await_complete:
            return operation
        return OperationWait(operation).completed

    # actually, half-async (because Requests)
    async def _async_instance_state_management(self, action=None, instance_id=None) -> CoroutineType:
        """Half async helper func for managing the state of the instance.
        Supported actions: start, restart, stop.
        Returns coroutine.
        """

        ACTIONS = ('start', 'restart', 'stop')
        if action not in ACTIONS:
            raise TypeError(f'Action {action} not supported')

        url = f'{self.compute_url}/compute/v1/instances/{instance_id}:{action}'
        response = self._request.post(url)
        operation = Operation.de_json(response, self)

        await OperationWait(operation).await_complete_async()

    def _instance_disk_management(self, instance_id: str, data=None,
                                  action=None, await_complete=True) -> Operation:
        """Helper func for managing the state of the disk.
        Supported actions: detachDisk, attachDisk.
        """

        ACTIONS = ('detachDisk', 'attachDisk')
        if action not in ACTIONS:
            raise TypeError(f'Action {action} not supported')

        url = f'{self.compute_url}/compute/v1/instances/{instance_id}:{action}'
        response = self._request.post(url, json=data)
        operation = Operation.de_json(response, self)
        if not await_complete:
            return operation
        return OperationWait(operation).completed

    # actually, half-async (because Requests)
    async def _async_instance_disk_management(self, instance_id: str,
                                              action=None, data=None) -> Operation:
        """Half async helper func for managing the state of the disk.
        Supported actions: detachDisk, attachDisk.
        Returns coroutine.
        """

        ACTIONS = ('detachDisk', 'attachDisk')
        if action not in ACTIONS:
            raise TypeError(f'Action {action} not supported')

        url = f'{self.compute_url}/compute/v1/instances/{instance_id}:{action}'
        response = self._request.post(url, json=data)
        operation = Operation.de_json(response, self)

        await OperationWait(operation).await_complete_async()

    def _convert_attached_disks(self, disks: list) -> Disk:
        """Helpers func for convert instance attached disks."""
        attached_disks = []
        for disk_id in disks:
            attached_disks.append(self.disk(disk_id, raw=True))
        return Disk.de_list(attached_disks, self)

    # Instance public methods

    @log
    def instance(self, instance_id: str, metadata=False) -> Instance:
        """Returns instance as object."""
        url = f'{self.compute_url}/compute/v1/instances/{instance_id}'
        if metadata:
            url += '?view=FULL'

        response = self._request.get(url)
        return Instance.de_json(response, self)

    @log
    def instances_in_folder(self, folder_id: str, page_size=1000, query_filter=None) -> [Instance]:
        """Returns list of instances in the folder."""
        url = f'{self.compute_url}/compute/v1/instances?folderId={folder_id}&pageSize={page_size}'
        if query_filter:
            url += f'&filter={query_filter}'
        response = self._request.get(url).get('instances')
        return Instance.de_list(response, self)

    @log
    def instance_serial_port_output(self, instance_id: str, port=1) -> str:
        """Returns instance serial port output as string."""
        url = f'{self.compute_url}/compute/v1/instances/{instance_id}:serialPortOutput?port={port}'
        response = self._request.get(url).get('contents')
        return response

    @log
    def instance_operations(self, instance_id: str, page_size=1000) -> [Operation]:
        """Returns list of instance operations."""
        url = f'{self.compute_url}/compute/v1/instances/{instance_id}/operations?pageSize={page_size}'
        response = self._request.get(url).get('operations')
        return Operation.de_list(response, self)

    @log
    def instance_attach_existent_disk(self, instance_id: str, disk_id: str, mode='READ_WRITE',
                                      device_name=None, auto_delete=False, await_complete=True,
                                      run_async_await=False) -> Operation:
        """Attach existent disk to the specified instance."""

        # prepare valid data for disk
        params = locals().copy()
        del params['self']
        data = {'attachedDiskSpec': AttachedDiskSpec.prepare(params, self)}

        if run_async_await:
            return self._async_instance_disk_management(instance_id,
                action='attachDisk', data=data)

        return self._instance_disk_management(instance_id, action='attachDisk',
            data=data, await_complete=await_complete)

    @log
    def instance_attach_new_disk(self, instance_id: str, size: int, mode='READ_WRITE',
                                 auto_delete=None, name=None, description=None, type_id=None,
                                 image_id=None, snapshot_id=None, device_name=None,
                                 await_complete=True, run_async_await=False) -> Operation:
        """Create new disk and attach him to the instance."""

        # prepare valid data for disk
        params = locals().copy()
        del params['self']
        data = {'attachedDiskSpec': AttachedDiskSpec.prepare(params, self)}

        if run_async_await:
            return self._async_instance_disk_management(instance_id,
                action='attachDisk', data=data)

        return self._instance_disk_management(instance_id, action='attachDisk',
            data=data, await_complete=await_complete)

    @log
    def instance_detach_disk(self, instance_id: str, disk_id=None, disk_name=None,
                             await_complete=True, run_async_await=False) -> Operation:
        """Detach disk from the instance."""
        if disk_id and disk_name:
            raise TooManyArguments('disk_id and disk_name received, but you can use only one param')

        if disk_id:
            data = {'diskId': disk_id}
        elif disk_name:
            data = {'diskName': disk_name}

        if run_async_await:
            return self._async_instance_disk_management(instance_id,
                action='detachDisk', data=data)

        return self._instance_disk_management(instance_id, action='detachDisk',
            data=data, await_complete=await_complete)

    @log
    def start_instance(self, instance_id: str, await_complete=True, run_async_await=False) -> Operation:
        """Start specified instance and return operation as object."""
        if run_async_await:
            return self._async_instance_state_management(action='start',
                instance_id=instance_id)

        return self._instance_state_management(action='start',
            instance_id=instance_id, await_complete=await_complete)

    @log
    def restart_instance(self, instance_id: str, await_complete=True, run_async_await=False) -> Operation:
        """Restart specified instance and return operation as object."""
        if run_async_await:
            return self._async_instance_state_management(action='restart',
                instance_id=instance_id)

        return self._instance_state_management(action='restart',
            instance_id=instance_id, await_complete=await_complete)

    @log
    def stop_instance(self, instance_id: str, await_complete=True, run_async_await=False) -> Operation:
        """Stop specified instance and return operation as object."""
        if run_async_await:
            return self._async_instance_state_management(action='stop',
                instance_id=instance_id)

        return self._instance_state_management(action='stop',
            instance_id=instance_id, await_complete=await_complete)

    @log
    def delete_instance(self, instance_id: str, await_complete=True, run_async_await=False) -> Operation:
        """Delete specified instance and return operation as object."""
        url = self.compute_url + f'/compute/v1/instances/{instance_id}'
        if run_async_await:
            return self._async_delete_resource(url)
        return self._delete_resource(url, await_complete=await_complete)

    @log
    # FIXME: make better
    def create_instance(self, yaml_spec, await_complete=True,
                        run_async_await=False) -> Operation:
        """
        This method provide basic interface for creating instance in Compute Cloud.

        Currently support only from yaml params.

        Args:
          :yaml_spec: str `/path/to/file.yaml`
          :await_complete: bool
          :run_async_await: bool

        Metadata must be string.
        See: https://cloud.yandex.ru/docs/compute/concepts/vm-metadata

        For user-data use this example:
        ```
        user-data: |
          #cloud-config
          users:
          - name: akimrx
            sudo: ALL=(ALL) NOPASSWD:ALL
            shell: /bin/bash
            ssh-authorized-keys:
              - ssh-rsa AAA...
        ```
        """

        url = f'{self.compute_url}/compute/v1/instances'
        data = convert_yaml_to_dict(yaml_spec)

        if run_async_await:
            return self._async_resource_create(url, data=data)
        return self._resource_create(url, data=data, await_complete=await_complete)

    @log
    def update_instance(self):
        raise MethodNotAvailable(f'Method update_instance is not support yet')

    @log
    def update_instance_metadata(self):
        raise MethodNotAvailable(f'Method update_instance_metadata is not support yet')

    # Disks public methods

    @log
    def disk(self, disk_id: str, raw=False) -> Disk:
        """Returns a disk as object."""
        url = f'{self.compute_url}/compute/v1/disks/{disk_id}'
        response = self._request.get(url)
        if raw:
            return response
        return Disk.de_json(response, self)

    @log
    def disk_operations(self, disk_id: str, page_size=1000) -> [Operation]:
        """Returns list of the disk operations."""
        url = f'{self.compute_url}/compute/v1/disks/{disk_id}/operations?pageSize={page_size}'
        response = self._request.get(url).get('operations')
        return Operation.de_list(response, self)

    @log
    def disks_in_folder(self, folder_id: str, page_size=1000, query_filter=None) -> [Disk]:
        """Return list of the disks in a folder."""
        url = f'{self.compute_url}/compute/v1/disks?folderId={folder_id}&pageSize={page_size}'
        if query_filter:
            url += f'&filter={query_filter}'
        response = self._request.get(url).get('disks')
        return Disk.de_list(response, self)

    @log
    def delete_disk(self, disk_id: str, await_complete=True, run_async_await=False) -> Operation:
        """Delete specified disk and returns operation as object."""
        url = f'{self.compute_url}/compute/v1/disks/{disk_id}'
        if run_async_await:
            return self._async_delete_resource(url)
        return self._delete_resource(url, await_complete=await_complete)

    @log
    def create_disk(self, folder_id: str, size: int, zone_id: str, name=None,
                    description=None, labels={}, type_id=None, image_id=None,
                    snapshot_id=None, await_complete=True, run_async_await=False) -> Operation:
        """Create disk and returns operation as object."""

        url = f'{self.compute_url}/compute/v1/disks'

        # Prepare valid data for new disk
        params = locals().copy()
        del params['self']
        data = DiskSpec.prepare(params, self)

        if run_async_await:
            return self._async_resource_create(url, data=data)
        return self._resource_create(url, data=data, await_complete=await_complete)

    @log
    def update_disk(self):
        raise MethodNotAvailable(f'Method update_disk is not support yet')

    # Snapshots public methods

    @log
    def snapshot(self, snapshot_id: str) -> Snapshot:
        """Returns the snapshot as object."""
        url = f'{self.compute_url}/compute/v1/snapshots/{snapshot_id}'
        response = self._request.get(url)
        return Snapshot.de_json(response, self)

    @log
    def snapshot_operations(self, snapshot_id: str, page_size=1000) -> [Operation]:
        """Returns list of the snapshot operations."""
        url = f'{self.compute_url}/compute/v1/snapshots/{snapshot_id}/operations?pageSize={page_size}'
        response = self._request.get(url).get('operations')
        return Operation.de_list(response, self)

    @log
    def snapshots_in_folder(self, folder_id: str, page_size=1000, query_filter=None) -> [Snapshot]:
        """Returns list of the snapshots in a folder."""
        url = f'{self.compute_url}/compute/v1/snapshots?folderId={folder_id}&pageSize={page_size}'
        if query_filter:
            url += f'&filter={query_filter}'
        response = self._request.get(url).get('snapshots')
        return Snapshot.de_list(response, self)

    @log
    def create_snapshot(self, folder_id: str, disk_id: str, name=None, description=None,
                        labels=None, await_complete=True, run_async_await=False) -> Operation:
        """Create the snapshot for specified disk and return operation as object."""
        url = f'{self.compute_url}/compute/v1/snapshots'

        params = locals().copy()
        del params['self']
        data = SnapshotSpec.prepare(params, self)

        if run_async_await:
            return self._async_resource_create(url, data=data)
        return self._resource_create(url, data=data, await_complete=await_complete)

    @log
    def delete_snapshot(self, snapshot_id: str, await_complete=True,
                        run_async_await=False) -> Operation:
        """Delete specified snapshot and returns operation as object."""
        url = f'{self.compute_url}/compute/v1/snapshots/{snapshot_id}'

        if run_async_await:
            return self._async_delete_resource(url)
        return self._delete_resource(url, await_complete=await_complete)

    @log
    def update_snapshot(self):
        raise MethodNotAvailable(f'Method update_snapshot is not support yet')


    # Aliases

    startInstance = start_instance
    stopInstance = stop_instance
    restartInstance = restart_instance
    createInstance = create_instance
    deleteInstance = delete_instance
    updateInstance = update_instance
    updateInstanceMetadata = update_instance_metadata
    instanceSerialPortOut = instance_serial_port_output
    folderInstances = instances_in_folder
    instanceOperations = instance_operations
    instanceAttachNewDisk = instance_attach_existent_disk
    instanceAttachExistentDisk = instance_attach_existent_disk
    instanceDetachDisk = instance_detach_disk
    diskOperations = disk_operations
    folderDisks = disks_in_folder
    deleteDisk = delete_disk
    createDisk = create_disk
    updateDisk = update_disk
    snapshotOperations = snapshot_operations
    folderSnapshots = snapshots_in_folder
    createSnapshot = create_snapshot
    deleteSnapshot = delete_snapshot
    updateSnapshot = update_snapshot


class CertificateClient(YandexCloudClient):
    """Yandex.Cloud Certificate Manager Client.

    Child class, which inherit the properties and methods
    from the YandexCloudClient parent class.

    Required Args:
      oauth_token: str
            or
      iam_token: str
            or
      service_account_key: dict

    Optional Args:
      request: Request
      base_url: str
      compute_url: str
      resource_manager_url: str
      iam_url: str
      operation_url: str
      certificate_url: str
      certificate_data_url: str
    """

    @log
    def certificate(self, certificate_id: str, view="FULL") -> Certificate:
        """Return certificate:
        BASIC - short info,
        FULL - full info with challenges.
        """
        url = f'{self.certificate_url}/certificate-manager/v1/certificates/{certificate_id}?view={view}'

        response = self._request.get(url)
        return Certificate.de_json(response, self)

    @log
    def certificate_content(self, certificate_id: str) -> CertificateContent:
        """Returns content (chain, key) for validated certificate."""
        url = f'{self.certificate_data_url}/certificate-manager/v1/certificates/{certificate_id}:getContent'

        response = self._request.get(url)
        return CertificateContent.de_json(response, self)

    @log
    def certificates_in_folder(self,
                               folder_id: str,
                               page_size=1000,
                               query_filter=None,
                               view="BASIC") -> [Certificate]:
        """Returns certificates in the specified folder:
        BASIC - short info,
        FULL - full info with challenges.
        """

        url = f'{self.certificate_url}/certificate-manager/v1/certificates?folderId={folder_id}&pageSize={page_size}&view={view}'
        if query_filter:
            url += f'&filter={query_filter}'
        response = self._request.get(url).get('certificates')
        return Certificate.de_list(response, self)

    @log
    def certificate_operations(self,
                               certificate_id: str,
                               page_size=1000) -> [Operation]:
        """Returns list of certificate operations."""
        url = f'{self.compute_url}/certificate-manager/v1/certificates{certificate_id}/operations?pageSize={page_size}'
        response = self._request.get(url).get('operations')
        return Operation.de_list(response, self)

    @log
    def create_user_certificate(self):
        raise MethodNotAvailable(f'Method create_user_certificate is not support yet')

    @log
    def request_new_letsencrypt_certificate(self,
                                            folder_id: str,
                                            name: str,
                                            domains: list,
                                            description: str = None,
                                            labels: dict = dict(),
                                            challenge_type: str = "DNS",
                                            await_complete=True,
                                            run_async_await=False) -> Operation:
        """Create new Let's encrypt free certificate in the specified folder.
        Challenge type must be DNS or HTTP.
        """

        url = f'{self.certificate_url}/certificate-manager/v1/certificates/requestNew'

        params = locals().copy()
        del params['self']
        data = CertificateRequestSpec.prepare(params, self)

        if run_async_await:
            return self._async_resource_create(url, data=data)
        return self._resource_create(url, data=data, await_complete=await_complete)

    @log
    def update_certificate(self,
                           certificate_id: str,
                           updateMask: str,
                           name=None,
                           description=None,
                           labels=None,
                           certificate=None,
                           chain=None,
                           private_key=None) -> Operation:
        raise MethodNotAvailable(f'Method update_certificate is not support yet')

    @log
    def delete_certificate(self,
                           certificate_id: str,
                           await_complete=True,
                           run_async_await=False) -> Operation:
        """Delete specified certificate and returns operation as object."""
        url = f'{self.certificate_url}/certificate-manager/v1/certificates/{certificate_id}'

        if run_async_await:
            return self._async_delete_resource(url)
        return self._delete_resource(url, await_complete=await_complete)


    # Aliases

    folderCertificates = certificates_in_folder
    certificateContent = certificate_content
    certificateOperations = certificate_operations
    createUserCertificate = create_user_certificate
    createLetsEncryptCertificate = request_new_letsencrypt_certificate
    updateCertificate = update_certificate
    deleteCertificate = delete_certificate



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains ApiKey, AccessKey and Key classes."""

from yandex_cloud_client.base import YandexCloudObject


class Key(YandexCloudObject):
    def __init__(self):
        pass


class ApiKey(YandexCloudObject):
    def __init__(self):
        pass


class AccessKey(YandexCloudObject):
    def __init__(self):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains Disk, AttachedDisk, AttachedDiskSpec and DiskSpec classes."""

import logging

from yandex_cloud_client.constants import MIN_DISK_SIZE, MAX_DISK_SIZE, DISK_MODES
from yandex_cloud_client.base import YandexCloudObject
from yandex_cloud_client.error import TooManyArguments
from yandex_cloud_client.utils.helpers import human_readable_size, string_to_datetime, disk_mode_converter

logger = logging.getLogger(__name__)


class Disk(YandexCloudObject):
    """This class representing a disk as an independent object.

    Different from AttachedDisk object.

    Attributes:
      :id: str
      :folder_id: str
      :created_at: datetime
      :name: str
      :description: str
      :labels: list
      :type_id: str
      :zone_id: str
      :size: int
      :product_ids: list
      :status: str
      :instance_ids: list
      :source_image_id: str or None
      :source_snapshot_id: str or None
      :client: object

    """

    def __init__(self,
                 id=None,
                 folder_id=None,
                 created_at=None,
                 name=None,
                 description=None,
                 labels=None,
                 type_id=None,
                 zone_id=None,
                 size=None,
                 product_ids=None,
                 status=None,
                 instance_ids=None,
                 source_image_id=None,
                 source_snapshot_id=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        # Required
        self.id = id

        # Optional
        self.folder_id = folder_id
        self.created_at = string_to_datetime(created_at) if created_at is not None else created_at
        self.name = name
        self.description = description
        self.labels = labels
        self.type_id = type_id
        self.zone_id = zone_id
        self.size = int(size) if size is not None else None
        self.product_ids = product_ids
        self.status = status
        self.instance_ids = instance_ids
        self.source_image_id = source_image_id
        self.source_snapshot_id = source_snapshot_id

        self.client = client
        self._id_attrs = (self.id,)

    @property
    def human_readable_size(self):
        """Converting bytes to human-readable size."""
        if self.size is not None:
            return human_readable_size(self.size)
        return self.size

    def delete(self, *args, **kwargs):
        """Shortcut client.delete_disk()."""
        return self.client.delete_disk(self.id, *args, **kwargs)

    def update(self, *args, **kwargs):
        """Shortcut for client.update_disk()."""
        pass

    def operations(self, page_size=1000, *args, **kwargs):
        """Shortcut for client.disk_operations()."""
        return self.client.disk_operations(self.id, page_size=page_size, *args, **kwargs)

    def snapshots(self, query_filter=None, *args, **kwargs):
        """Shortcut for client.snapshots_in_folder()."""
        result = list()
        snapshots = self.client.snapshots_in_folder(self.folder_id, page_size=1000,
            query_filter=query_filter, *args, **kwargs)
        for snapshot in snapshots:
            if self.id == snapshot.source_disk_id:
                result.append(snapshot)
        return result

    def create_snapshot(self, name=None, description=None, labels=None, await_complete=True,
                        run_async_await=False, *args, **kwargs):
        """Shortcut for client.create_snapshot()."""
        return self.client.create_snapshot(self.folder_id, self.id, name=name, description=description, labels=labels,
            await_complete=await_complete, run_async_await=run_async_await, *args, **kwargs)

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(Disk, cls).de_json(data, client)
        return cls(client=client, **data)

    @classmethod
    def de_list(cls, data: list, client):
        if not data:
            return []

        disks = list()
        for disk in data:
            disks.append(cls.de_json(disk, client))

        return disks


class AttachedDisk(YandexCloudObject):
    """This object represents a attached to the instance disk.

    Different from a Disk object.

    Attributes:
      :id: str
      :attached_to: str
      :mode: str
      :device_name: str
      :auto_delete: bool
      :folder_id: str
      :client: object

    """

    def __init__(self,
                 disk_id=None,
                 attached_to=None,
                 mode=None,
                 device_name=None,
                 auto_delete=None,
                 folder_id=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        # Required
        self.id = disk_id  # alias for disk_id
        self.disk_id = disk_id  # stay original attr

        # Optional
        self.folder_id = folder_id
        self.attached_to = attached_to
        self.mode = mode
        self.device_name = device_name
        self.auto_delete = auto_delete

        self.client = client
        self._id_attrs = (self.id,)

    def detach(self, await_complete=True, run_async_await=False, *args, **kwargs):
        """Shortcut for client.instance_detach_disk()."""
        return self.client.instance_detach_disk(self.attached_to, disk_id=self.id,
            await_complete=await_complete, run_async_await=run_async_await, *args, **kwargs)

    def delete(self, *args, **kwargs):
        """Shortcut for client.delete_disk()."""
        return self.client.delete_disk(self.id, *args, **kwargs)

    def update(self, *args, **kwargs):
        """Shortcut for client.update_disk()."""
        pass

    def operations(self, page_size=1000, *args, **kwargs):
        """Shortcut for client.disk_operarions()."""
        return self.client.disk_operations(self.id, page_size=page_size, *args, **kwargs)

    def snapshots(self, query_filter=None, *args, **kwargs):
        """Shortcut for client.snapshots_in_folder()."""
        result = list()
        snapshots = self.client.snapshots_in_folder(self.folder_id, page_size=1000,
            query_filter=query_filter, *args, **kwargs)
        for snapshot in snapshots:
            if self.id == snapshot.source_disk_id:
                result.append(snapshot)
        return result

    def create_snapshot(self, name=None, description=None, labels=None, await_complete=True,
                        run_async_await=False, *args, **kwargs):
        """Shortcut for client.create_snapshot()."""
        return self.client.create_snapshot(self.folder_id, self.id, name=name, description=description, labels=labels,
            await_complete=await_complete, run_async_await=run_async_await, *args, **kwargs)

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(AttachedDisk, cls).de_json(data, client)
        return cls(client=client, **data)

    @classmethod
    def de_list(cls, data: list, client):
        if not data:
            return []

        disks = list()
        for disk in data:
            disks.append(cls.de_json(disk, client))

        return disks


# Specifications for disks


class DiskSpec(YandexCloudObject):
    """This object represents new disk."""

    def __init__(self,
                 folder_id=None,
                 size=None,
                 zone_id=None,
                 name=None,
                 description=None,
                 labels=None,
                 type_id=None,
                 image_id=None,
                 snapshot_id=None,
                 client=None,
                 **kwargs):

        # Required
        self.folderId = folder_id
        self.size = int(size) if size is not None else size
        self.zoneId = zone_id

        # Optional
        self.name = name
        self.description = description
        self.labels = labels
        self.typeId = type_id
        self.imageId = image_id
        self.snapshotId = snapshot_id

        self.client = client
        self._validate()

    def _validate(self):
        if self.imageId and self.snapshotId:
            raise TooManyArguments('snapshot_id and image_id received, but you can use only one param')

        if self.size < MIN_DISK_SIZE or self.size > MAX_DISK_SIZE:
            raise ValueError(f'The disk size must be between {MIN_DISK_SIZE} and {MAX_DISK_SIZE} bytes')

    @classmethod
    def prepare(cls, data: dict, client):
        """Deserializing and preparing for a request."""
        if not data:
            return None

        data = super(DiskSpec, cls).de_json(data, client)
        result = cls(client=client, **data).to_dict()

        # cleaning None-type keys
        cleaner = lambda x: dict((k, v) for (k, v) in x.items() if v is not None)

        return cleaner(result)


class AttachedDiskSpec(YandexCloudObject):
    """This object represents existent or new disk,
    which will be immediately connected to the instance.

    """

    def __init__(self,
                 disk_id=None,
                 device_name=None,
                 mode=None,
                 auto_delete=None,
                 size=None,
                 name=None,
                 description=None,
                 labels=None,
                 type_id=None,
                 image_id=None,
                 snapshot_id=None,
                 client=None,
                 **kwargs):

        # Required for existent
        self.diskId = disk_id

        # Optional general
        self.deviceName = device_name
        self.mode = disk_mode_converter(mode) if mode is not None else mode
        self.autoDelete = auto_delete

        # Required for new
        self.size = int(size) if size is not None else size

        # Optional for new
        self.name = name
        self.description = description
        self.labels = labels
        self.typeId = type_id
        self.imageId = image_id
        self.snapshotId = snapshot_id

        self.client = client
        self._validate()

    def _validate(self):
        if self.imageId and self.snapshotId:
            raise TooManyArguments('snapshot_id and image_id received, but you can use only one param')

        if self.mode is not None and self.mode not in DISK_MODES:
            raise ValueError(f'Invalid disk mode: {self.mode}. Supported modes: {DISK_MODES}')

        if self.size is not None:
            if self.size < MIN_DISK_SIZE or self.size > MAX_DISK_SIZE:
                raise ValueError(f'The disk size must be between {MIN_DISK_SIZE} and {MAX_DISK_SIZE} bytes')

    @classmethod
    def prepare(cls, data: dict, client):
        """Deserializing and preparing for a request."""
        if not data:
            return None

        new = ('name', 'size', 'typeId', 'description', 'imageId', 'snapshotId')
        data = super(AttachedDiskSpec, cls).de_json(data, client)
        mdata = cls(client=client, **data).to_dict()

        # cleaning None-type keys
        cleaner = lambda x: dict((k, v) for (k, v) in x.items() if k not in new and v is not None)
        # separate and prepare diskSpec for new disk
        new_disk = lambda x: dict((k, v) for (k, v) in x.items() if k in new and v is not None)

        result = cleaner(mdata)
        result.update({'diskSpec': new_disk(mdata)})

        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains Request wrapper."""

import re
import json
import logging
import requests

from enum import Enum

from yandex_cloud_client.constants import DEFAULT_TIMEOUT
from yandex_cloud_client.utils.response import Response
from yandex_cloud_client.utils.decorators import retry

from yandex_cloud_client.error import (Unauthorized, BadRequest, PermissionDenied, NetworkError,
                                       YandexCloudError, TimedOut, ResourceNotFound,
                                       FeatureNotImplemented, ReourceExhausted, HTTPError)


logger = logging.getLogger('requests.packages.urllib3')

HEADERS = {
    'content-type': 'application/json'
}


class Request:
    """Base HTTP request wrapper for Yandex.Cloud API.

    If you need proxy, use:
    proxy_url='socks5://user:password@host:port'

    """

    def __init__(self,
                 client=None,
                 headers=None,
                 proxy_url=None,
                 timeout=5):

        self.headers = headers or HEADERS.copy()
        self.client = self.set_and_return_client(client)
        self.proxies = {'http': proxy_url, 'https': proxy_url} if proxy_url else None
        self.timeout = int(timeout) if timeout is not None else DEFAULT_TIMEOUT

    def set_authorization(self, token):
        self.headers.update({'Authorization': f'Bearer {token}'})

    def set_and_return_client(self, client):
        self.client = client

        if self.client and self.client.token:
            self.set_authorization(self.client.token)

        return self.client

    @staticmethod
    def _convert_camel_to_snake(text):
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', text)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

    @staticmethod
    def _object_hook(obj: dict):
        cleaned_object = {}
        for key, value in obj.items():
            key = Request._convert_camel_to_snake(key.replace('-', '_'))

            if len(key) and key[0].isdigit():
                key = '_' + key

            cleaned_object.update({key: value})
        return cleaned_object

    def _parse(self, json_data: bytes):
        try:
            decoded_s = json_data.decode('utf-8')
            data = json.loads(decoded_s, object_hook=Request._object_hook)
        except UnicodeDecodeError:
            logger.debug('Logging raw invalid UTF-8 response:\n%r', json_data)
            raise YandexCloudError('Server response could not be decoded using UTF-8')
        except (AttributeError, ValueError):
            raise YandexCloudError('Invalid server response', json_data)

        if data.get('result') is None:
            data = {
                'result': data,
                'error': data.get('error') or RpcError(data.get('code', 2)),
                'error_description': data.get('error_description') or data.get('message')
            }

        return Response.de_json(data, self.client)

    @retry((NetworkError, TimedOut))
    def _request_wrapper(self, *args, **kwargs):
        if 'headers' not in kwargs:
            kwargs['headers'] = {}

        try:
            resp = requests.request(*args, **kwargs)
        except requests.Timeout:
            raise TimedOut()
        except requests.RequestException as e:
            raise NetworkError(e)

        if 200 <= resp.status_code <= 299:
            return resp

        logger.debug(resp.content)
        parse = self._parse(resp.content)
        message = parse.error or 'Unknown HTTPError'

        if resp.status_code == 401:
            raise Unauthorized(message)
        elif resp.status_code == 400:
            raise BadRequest(message)
        elif resp.status_code == 403:
            raise PermissionDenied(message)
        elif resp.status_code == 404:
            raise ResourceNotFound(message)
        elif resp.status_code in (409, 413):
            raise HTTPError(f'HTTP {resp.status_code} – {message}')
        elif resp.status_code == 429:
            raise ReourceExhausted(message)

        elif resp.status_code == 510:
            raise FeatureNotImplemented(message)
        else:
            raise HTTPError(f'{resp.status_code} – {message}')

    def get(self, url, params=None, *args, **kwargs):
        result = self._request_wrapper('GET', url, params=params, headers=self.headers,
            proxies=self.proxies, timeout=self.timeout, *args, **kwargs)

        return self._parse(result.content).result

    def post(self, url, data=None, json=None, *args, **kwargs):
        result = self._request_wrapper('POST', url, headers=self.headers, proxies=self.proxies,
            data=data, json=json, timeout=self.timeout, *args, **kwargs)

        return self._parse(result.content).result

    def put(self, url, data=None, json=None, *args, **kwargs):
        result = self._request_wrapper('PUT', url, headers=self.headers, proxies=self.proxies,
            data=data, json=json, timeout=self.timeout, *args, **kwargs)

        return self._parse(result.content).result

    def patch(self, url, data=None, json=None, *args, **kwargs):
        result = self._request_wrapper('PATCH', url, headers=self.headers, proxies=self.proxies,
            data=data, json=json, timeout=self.timeout, *args, **kwargs)

        return self._parse(result.content).result

    def delete(self, url, *args, **kwargs):
        result = self._request_wrapper('DELETE', url, headers=self.headers, proxies=self.proxies,
            timeout=self.timeout, *args, **kwargs)

        return self._parse(result.content).result


class RpcError(Enum):
    """This class convert grpc digit codes to messages."""

    CANCELLED = 1
    UNKNOWN = 2
    INVALID_ARGUMENT = 3
    DEADLINE_EXCEEDED = 4
    NOT_FOUND = 5
    ALREADY_EXISTS = 6
    PERMISSION_DENIED = 7
    RESOURCE_EXHAUSTED = 8
    FAILED_PRECONDITION = 9
    ABORTED = 10
    OUT_OF_RANGE = 11
    NOT_IMPLEMENTED = 12
    INTERNAL = 13
    UNAVAILABLE = 14
    DATA_LOSS = 15
    UNAUTHENTICATED = 16



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains Instance, Metadata, SchedulingPolicy, Resources classes."""

import logging

from yandex_cloud_client.utils.helpers import universal_obj_hook
from yandex_cloud_client.base import YandexCloudObject
from yandex_cloud_client.vpc.address import Address

logger = logging.getLogger(__name__)


class NetworkInterface(YandexCloudObject):
    """This object represents a network interfaces of instance."""

    def __init__(self,
                 index=None,
                 mac_address=None,
                 subnet_id=None,
                 primary_v4_address=None,
                 primary_v6_address=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.index = index
        self.mac_address = mac_address
        self.subnet_id = subnet_id
        self.primary_v4_address = primary_v4_address
        self.primary_v6_address = primary_v6_address

        self.client = client
        self._id_attrs = (self.index, self.mac_address,)

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(NetworkInterface, cls).de_json(data, client)
        data['primary_v4_address'] = Address.de_json(universal_obj_hook(data.get('primary_v4_address')), client)
        data['primary_v6_address'] = Address.de_json(universal_obj_hook(data.get('primary_v6_address')), client)
        return cls(client=client, **data)

    @classmethod
    def de_list(cls, data: list, client):
        if not data:
            return []

        interfaces = list()
        for interface in data:
            interfaces.append(cls.de_json(interface, client))

        return interfaces



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains decorator functions."""

import time
import logging
from functools import wraps
from decorator import decorate

logger = logging.getLogger(__name__)


def retry(exceptions, tries=4, delay=5, backoff=2, logs=True):
    """Network errors decorator for retry requests."""
    def retry_decorator(func):
        @wraps(func)
        def func_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    msg = f'Retrying in {mdelay} seconds.. caused by NetworkError: {e}'
                    if logs:
                        logger.warning(msg)
                    else:
                        print(msg)
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return func(*args, **kwargs)
        return func_retry
    return retry_decorator


def log(func, *args, **kwargs):
    """Add debug messages to logger."""
    logger = logging.getLogger(func.__module__)

    def decorator(self, *args, **kwargs):
        logger.debug(f'Entering {func.__name__}')
        result = func(*args, **kwargs)
        logger.debug(result)
        logger.debug(f'Exiting {func.__name__}')
        return result

    return decorate(func, decorator)



#!/usr/bin/env python3
"""This module contains Certificate, CertificateContent, CertificateRequestSpec, Challenges, DnsChallenges and HttpChallenges classes."""

import logging

from yandex_cloud_client.base import YandexCloudObject
from yandex_cloud_client.utils.helpers import universal_obj_hook, string_to_datetime

logger = logging.getLogger(__name__)


class Certificate(YandexCloudObject):
    """This class representing a managed certificate.

    Attributes:
      :id: str
      :folder_id: str
      :created_at: datetime
      :name: str
      :desription: str
      :labels: dict
      :type: str
      :domains: list
      :status: str
      :subject: str
      :serial: str
      :updated_at: datetime
      :issued_at: datetime
      :not_after: datetime
      :not_before: datetime
      :challenges: list of objects

    """

    def __init__(self,
                 id=None,
                 folder_id=None,
                 created_at=None,
                 name=None,
                 description=None,
                 labels=None,
                 type=None,
                 domains=None,
                 status=None,
                 subject=None,
                 serial=None,
                 updated_at=None,
                 issued_at=None,
                 not_after=None,
                 not_before=None,
                 challenges=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        # Required
        self.id = id

        # Optional
        self.folder_id = folder_id
        self.created_at = string_to_datetime(created_at) if created_at is not None else created_at
        self.name = name
        self.description = description
        self.labels = labels
        self.type = type
        self.domains = domains
        self.status = status
        self.subject = subject
        self.serial = serial
        self.updated_at = string_to_datetime(updated_at) if updated_at is not None else updated_at
        self.issued_at = string_to_datetime(issued_at) if issued_at is not None else issued_at
        self.not_after = string_to_datetime(not_after) if not_after is not None else not_after
        self.not_before = string_to_datetime(not_before) if not_before is not None else not_before
        self.challenges = challenges

        self.client = client
        self._id_attrs = (self.id,)

    @property
    def expires(self):
        return self.not_after

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(Certificate, cls).de_json(data, client)
        data['challenges'] = Challenges.de_list(universal_obj_hook(data.get('challenges')), client)
        return cls(client=client, **data)

    @classmethod
    def de_list(cls, data: list, client):
        if not data:
            return []

        certificates = list()
        for certificate in data:
            certificates.append(cls.de_json(certificate, client))

        return certificates

    def content(self, *args, **kwargs):
        """Shortcut for client.certificate_content()."""
        return self.client.certificate_content(self.id, *args, **kwargs)

    def delete(self, run_async_await=False, await_complete=True, *args, **kwargs):
        """Shortcut for client.delete_certificate()."""
        return self.client.delete_certificate(self.id, await_complete, run_async_await, *args, **kwargs)


class Challenges(YandexCloudObject):
    """This object represents the certificate challenges."""

    def __init__(self,
                 domain=None,
                 type=None,
                 created_at=None,
                 updated_at=None,
                 status=None,
                 message=None,
                 error=None,
                 dns_challenge=None,
                 http_challenge=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.domain = domain
        self.type = type
        self.created_at = string_to_datetime(created_at) if created_at is not None else created_at
        self.updated_at = string_to_datetime(updated_at) if updated_at is not None else updated_at
        self.status = status
        self.message = message
        self.error = error
        self.dns_challenge = dns_challenge
        self.http_challenge = http_challenge

        self.client = client

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(Challenges, cls).de_json(data, client)

        data['dns_challenge'] = DnsChallenges.de_json(universal_obj_hook(data.get('dns_challenge')), client)
        data['http_challenge'] = HttpChallenges.de_json(universal_obj_hook(data.get('http_challenge')), client)

        return cls(client=client, **data)

    @classmethod
    def de_list(cls, data: list, client):
        if not data:
            return []

        challenges = list()
        for challenge in data:
            challenges.append(cls.de_json(challenge, client))

        return challenges


class DnsChallenges(YandexCloudObject):
    """This object represents a DNS Challenge for the certificate."""

    def __init__(self,
                 name=None,
                 type=None,
                 value=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.name = name
        self.type = type
        self.value = value

        self.client = client

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(DnsChallenges, cls).de_json(data, client)
        return cls(client=client, **data)


class HttpChallenges(YandexCloudObject):
    """This object represents a HTTP Challenge for the certificate."""

    def __init__(self,
                 url=None,
                 content=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.url = url
        self.content = content

        self.client = client

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(HttpChallenges, cls).de_json(data, client)
        return cls(client=client, **data)


class CertificateContent(YandexCloudObject):
    """This object represents a content for the certificate."""

    def __init__(self,
                 id=None,
                 certificate_id=None,
                 certificate_chain=None,
                 private_key=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.id = certificate_id
        self.certificate_id = certificate_id
        self.certificate_chain = certificate_chain
        self.private_key = private_key

        self.client = client
        self._id_attrs = (self.id,)

    @property
    def chain(self):
        return self.certificate_chain

    @property
    def fullchain(self):
        return self.certificate_chain

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(CertificateContent, cls).de_json(data, client)
        return cls(client=client, **data)


class CertificateRequestSpec(YandexCloudObject):
    """This object represents a specification for new certificate."""

    def __init__(self,
                 folder_id=None,
                 name=None,
                 description=None,
                 labels=None,
                 domains=None,
                 challenge_type=None,
                 client=None,
                 **kwargs):

        # Required
        self.folderId = folder_id

        # Optional
        self.name = name
        self.description = description
        self.labels = labels
        self.domains = domains
        self.challengeType = challenge_type

        self.client = client

    @classmethod
    def prepare(cls, data: dict, client):
        """Deserializing and preparing for a request."""
        if not data:
            return None

        data = super(CertificateRequestSpec, cls).de_json(data, client)
        result = cls(client=client, **data).to_dict()

        # cleaning None-type keys
        cleaner = lambda x: dict((k, v) for (k, v) in x.items() if v is not None)

        return cleaner(result)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains YandexCloudObject parent class."""

import json
import logging

from abc import ABCMeta

logger = logging.getLogger(__name__)


class YandexCloudObject(object):
    """Base class for Yandex.Cloud objects."""

    __metaclass__ = ABCMeta
    _id_attrs = ()

    def __str__(self):
        return str(self.to_dict())

    def __repr__(self):
        return str(self)

    def __getitem__(self, item):
        return self.__dict__[item]

    @staticmethod
    def handle_unknown_kwargs(obj, **kwargs):
        """Logging unparsed fields from Yandex.Cloud API."""
        if kwargs:
            logger.debug(f'Unparsed fields received from API')
            logger.debug(f'Type: {type(obj)}; kwargs: {kwargs}')

    @classmethod
    def de_json(cls, data: dict, client):
        """Deserialize object."""
        if not data:
            return None

        data = data.copy()
        return data

    def to_json(self):
        """Serialize object to json."""
        return json.dumps(self.to_dict())

    def to_dict(self) -> dict:
        """Recursive serialize object."""
        def parse(val):
            if hasattr(val, 'to_dict'):
                return val.to_dict()
            elif isinstance(val, list):
                return [parse(it) for it in val]
            elif isinstance(val, dict):
                return {key: parse(value) for key, value in val.items()}
            else:
                return val

        data = self.__dict__.copy()
        data.pop('client', None)
        data.pop('_id_attrs', None)

        return parse(data)

    def to_clean_dict(self) -> dict:
        """Delete None-type key: value pairs."""

        data = self.to_dict().copy()
        cleaner = lambda x: dict((k, v) for (k, v) in x.items() if v is not None)
        return cleaner(data)

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return self._id_attrs == other._id_attrs
        return super(YandexCloudObject, self).__eq__(other)

    def __hash__(self) -> int:
        if self._id_attrs:
            return hash((self.__class__, self._id_attrs))
        return super(YandexCloudObject, self).__hash__()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages


def readme():
    with open("README.md", "r") as fh:
        long_description = fh.read()

    return long_description


def requirements():
    requirements_list = []

    with open('requirements.txt') as requirements:
        for install in requirements:
            requirements_list.append(install.strip())

    return requirements_list


packages = find_packages()
requirements = requirements()


setup(
    name='yandex-cloud-client',
    version='1.0.2b',
    author='Akim Faskhutdinov',
    author_email='akimstrong@yandex.ru',
    license='GPLv3',
    description='Unofficial Yandex.Cloud REST API Client',
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/akimrx/python-yc-client",
    keywords='yandex cloud rest api client',
    packages=packages,
    install_requrements=requirements,
    include_package_data=True,
    python_requires='>=3.6',
    zip_safe=False
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains an object that represents Yandex.Cloud errors."""

class YandexCloudError(Exception):
    pass


class InvalidToken(YandexCloudError):
    pass


class Unauthorized(YandexCloudError):
    pass


class NetworkError(YandexCloudError):
    pass


class BadRequest(YandexCloudError):
    pass


class ResourceNotFound(YandexCloudError):
    pass


class TooManyArguments(YandexCloudError):
    pass


class CertificateError(YandexCloudError):
    pass


class PermissionDenied(YandexCloudError):
    pass


class MethodNotAvailable(YandexCloudError):
    pass


class FeatureNotImplemented(YandexCloudError):
    pass


class ReourceExhausted(YandexCloudError):
    pass


class OperationDeadlineExceeded(YandexCloudError):
    pass


class HTTPError(YandexCloudError):
    pass


class TimedOut(YandexCloudError):
    def __init__(self):
        super().__init__('Timed out')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains Response class."""

from yandex_cloud_client.base import YandexCloudObject


class Response(YandexCloudObject):
    """This object represents a response from Yandex.Cloud API."""

    def __init__(self,
                 data: dict,
                 result=None,
                 error=None,
                 error_description=None,
                 client=None,
                 **kwargs):

        super().handle_unknown_kwargs(self, **kwargs)

        self.data = data
        self._result = result
        self._error = error
        self.error_description = error_description

        self.client = client

    @property
    def error(self) -> str:
        return f'{self._error}: {self.error_description if self.error_description else ""}'

    @property
    def result(self) -> dict:
        return self.data if self._result is None else self._result

    @classmethod
    def de_json(cls, data: dict, client):
        if not data:
            return None

        data = super(Response, cls).de_json(data, client)
        data['data'] = data.copy()
        return cls(client=client, **data)
